// By SeungWon Lee - minor modification applied
//
// 2016.05.16 - Translate comment to english from deutsches
// 2016.05.01 - Add interrupt handler handling logic and function for CAN RX
//
// Original source code: http://www.mikrocontroller.net/topic/98697 mod version of Michael

#include "can.h"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>  
#include <inttypes.h>


int8_t iemob[NUMBER_OF_MOBS] = {
	IEMOB0, IEMOB1, IEMOB2,  IEMOB3,  IEMOB4,  IEMOB5,  IEMOB6, IEMOB7,
	IEMOB8, IEMOB9, IEMOB10, IEMOB11, IEMOB12, IEMOB13, IEMOB14 };

// extended id
bool extended_id;

/*******************************************************************************
 Activate CAN controller

 Parameter:
    baudrate: 0,1,2...5 (for 100, 125, 200, 250, 500, 1000)
    intmode: event in which an interrupt is to be triggered
     - NONE  - Deactivated
     - TX    - Sending data
     - RX    - Received data
     - TXRX  - Data sent and / or received
	extendedId: Extended CAN or not
*******************************************************************************/
int can_init(CanBaudRate baudrate, CanInterruptMode intmode, bool extendedId)
{
	uint8_t i;
	extended_id = extendedId;

	// Initialize status and control registers of all message objects
	for (i = 0; i < NUMBER_OF_MOBS; i++) {
		can_get_mob(i);
		CANSTMOB = 0;
		CANCDMOB = 0;
	}

	// Set baudrate
	if (can_set_baudrate(baudrate) == 0) {
		return 0;
	}
	// Set interrupt mode
	if (can_set_interrupt(intmode) == 0) {
		return 0;
	}

	// Set CAN controller in Enable mode
	setbit(CANGCON, ENASTB);

	//Wait until the CAN controller set the Enabled bit and is ready for use
	while (!getbit(CANGSTA, ENFG));

	return 1; 
}

/*******************************************************************************
 Deactivate CAN controller
*******************************************************************************/
int can_deinit(void)
{
	uint8_t i;
	for (i = 0; i < 15; i++) {
		can_disable_mob(i);
	}

	// Turn off CAN controller
	clearbit(CANGCON, ENASTB);

	// Wait until really off
	while (getbit(CANGSTA, ENFG));

	return 1;
}

/*******************************************************************************
 Set auto replay
*******************************************************************************/
void can_set_autoreply(uint8_t mob, uint8_t length, uint8_t *buf)
{
	can_get_mob(mob);
	can_set_mode(CAN_MODE_AUTO_REPLY);
	can_set_data(length, buf);
}

/*******************************************************************************
 Set CAN baudrate
 16 MHz only
*******************************************************************************/
int can_set_baudrate(CanBaudRate baudrate)
{
	typedef struct {
		uint8_t BT1;
		uint8_t BT2;
		uint8_t BT3;
	} SetCANBT;
	const SetCANBT CANBaudrateSet[CAN_BAUDRATE_1000K + 1] = {
		{ 0x12, 0x0c, 0x37 },	// CAN_BAUDRATE_100K
		{ 0x0e, 0x0c, 0x37 },	// CAN_BAUDRATE_125K
		{ 0x08, 0x0c, 0x37 },	// CAN_BAUDRATE_200K
		{ 0x06, 0x0c, 0x37 },	// CAN_BAUDRATE_250K
		{ 0x02, 0x0c, 0x37 },	// CAN_BAUDRATE_500K
		{ 0x00, 0x0c, 0x37 }	// CAN_BAUDRATE_1000K
	};
	
	if (baudrate <= CAN_BAUDRATE_1000K) {
		CANBT1 = CANBaudrateSet[baudrate].BT1;
		CANBT2 = CANBaudrateSet[baudrate].BT2;
		CANBT3 = CANBaudrateSet[baudrate].BT3;
		return 1;
	}
	return 0;
}

/*******************************************************************************
Function sets the operating mode of the previously selected object

Parameter:
	uint8_t mob: Issue to be selected object (0-14)

	uint8_t mode
	- Mode of the Message object:
	- DISABLED Deactivated
	- Send TRANSMIT_DATA data
	- Send TRANSMIT_REMOTE request
	- RECEIVE_DATA receive mode
	- AUTO_REPLY automatic response mode

*******************************************************************************/
int can_enable_mob(uint8_t mob, uint8_t mode, uint32_t id, uint32_t idm)
{
	// Select object
	can_get_mob(mob);

	// Enable interrupt for this property
	can_set_mob_interrupt(mob);

	// Set ID mask
	can_set_id_mask(idm);

	// Set ID
	can_set_id(id);

	// Set operating mode
	return can_set_mode(mode);
}

/*******************************************************************************
*******************************************************************************/
int can_send_data(uint8_t mob, uint8_t length, uint8_t * data)
{
	uint8_t mode;

	// Select object
	can_get_mob(mob);
  
	// Save current operating mode
	mode = can_get_mode();

	// Write user data in registry
	can_set_data(length, data);

	// Start data transfer
	can_set_mode(CAN_MODE_TRANSMIT_DATA);

	// Wait is completed until the data transfer (Set TXOK flag of CAN controller)
	while (!getbit(CANSTMOB, TXOK));      

	// Clear TXOK flag by hand
	clearbit(CANSTMOB, TXOK);

	// Restore Old mode
	can_set_mode(mode);

	return 1;
}


/*******************************************************************************
 Set interrupt mode
 
    mode: event in which an interrupt is to be triggered
    - NONE  - Deactivated
    - TX    - Sending data
    - RX    - Received data
    - TXRX  - Data sent and / or received
*******************************************************************************/
int can_set_interrupt(CanInterruptMode mode)
{
	switch(mode) {
		case CAN_INTERRUPT_NONE:
			clearbit(CANGIE, ENIT);
			clearbit(CANGIE, ENRX);
			clearbit(CANGIE, ENTX);
			break;

		case CAN_INTERRUPT_TX:
			setbit(CANGIE, ENIT);
			clearbit(CANGIE, ENRX);
			setbit(CANGIE, ENTX);
			break;

		case CAN_INTERRUPT_RX:
			setbit(CANGIE, ENIT);
			setbit(CANGIE, ENRX);
			clearbit(CANGIE, ENTX);
			break;

		case CAN_INTERRUPT_TXRX:
			setbit(CANGIE, ENIT);
			setbit(CANGIE, ENRX);
			setbit(CANGIE, ENTX);
			break;

		default:
			return 0;
	}

	return 1;
}

/*******************************************************************************
 Selects CanPage of the object and makes access to registers of the object
*******************************************************************************/
void can_get_mob(uint8_t mob)
{
	CANPAGE = (mob << 4);
}
  
/*******************************************************************************
 Sets ID mask of an object to a new value.
	Parameters: uint32_t idm: ID mask in decimal
	while the values of the lower 3 bits (RTRTAG, Reserved and IDEMSK) remain in CANIDM4.
*******************************************************************************/
void can_set_id_mask(uint32_t idm)
{
	// Standard identifier (11 bit)
	if (!extended_id)
	{
		CANIDM2 = (uint8_t)(idm << 5); 
		CANIDM1 = (uint8_t)(idm >> 3);
	}
	// Extended identifier
	else
	{
		idm <<= 3;
		idm  |= 7;

		CANIDM4 = (int8_t) (idm);
		CANIDM3 = (int8_t) (idm>>8);
		CANIDM2 = (int8_t) (idm>>16);
		CANIDM1 = (int8_t) (idm>>24);
	}
}

/*******************************************************************************
 Fetches ID of the received message
*******************************************************************************/
uint32_t can_get_id(void)
{
	uint32_t id = 0;

	// Standard identifier (11 bit)
	if (!extended_id)
	{
		id  = (uint8_t)  CANIDT2 >> 5;
		id |= (uint16_t) CANIDT1 << 3;
	}
	// Extended identifier
	else
	{
		id |= ((uint32_t) CANIDT1 << 24);
		id |= ((uint32_t) CANIDT2 << 16);
		id |= ((uint32_t) CANIDT3 << 8);
		id |= (CANIDT4&0xF8);
		id >>= 3;
	}

	return id;
}

/*******************************************************************************
 Sets ID of an object to a new value. 
 while the values of the lower 3 bits (RTRTAG, RB1TAG and RB0TAG) remain in CANIDM4.
*******************************************************************************/
void can_set_id(uint32_t id)
{
	// Standard identifier (11 bit)
	if (!extended_id)
	{
		CANIDT2 = (uint8_t)(id << 5); 
		CANIDT1 = (uint8_t)(id >> 3);
	}
	// Extended identifier
	else
	{
		id <<= 3;
		id &= 0xfffffff8;
		id |= (CANIDT4 & 0x07);

		CANIDT4 = (int8_t) (id);
		CANIDT3 = (int8_t) (id>>8);
		CANIDT2 = (int8_t) (id>>16);
		CANIDT1 = (int8_t) (id>>24);
	}
}


/*******************************************************************************
 Sets the operating mode of the previously selected object.
 Parameters: uint8_t mode:
 - Mode of the Message object:
 - DISABLED			Deactivated
 - TRANSMIT_DATA	Send data
 - TRANSMIT_REMOTE	Send inquiry
 - RECEIVE_DATA		Receive mode
 - AUTO_REPLY		Automatic answer mode
*******************************************************************************/
int can_set_mode(uint8_t mode)
{
	if (extended_id)
		setbit(CANCDMOB, IDE);
	else
		clearbit(CANCDMOB, IDE);

	switch(mode) {
		case CAN_MODE_DISABLED:
			clearbit(CANCDMOB, CONMOB0);
			clearbit(CANCDMOB, CONMOB1);
			clearbit(CANCDMOB, RPLV);
			clearbit(CANIDT4, RTRTAG);
			clearbit(CANIDM4, RTRMSK);
			break;

		case CAN_MODE_TRANSMIT_DATA:
			setbit(CANCDMOB, CONMOB0);
			clearbit(CANCDMOB, CONMOB1);
			clearbit(CANCDMOB, RPLV);
			clearbit(CANIDT4, RTRTAG);
			break;

		case CAN_MODE_TRANSMIT_REMOTE:
			clearbit(CANCDMOB, CONMOB1);
			setbit(CANCDMOB, CONMOB0);
			clearbit(CANCDMOB, RPLV);
			setbit(CANIDT4, RTRTAG);
			break;

		case CAN_MODE_RECEIVE_DATA:
			clearbit(CANCDMOB, CONMOB0);
			setbit(CANCDMOB, CONMOB1);
			clearbit(CANCDMOB, RPLV);
			clearbit(CANIDT4, RTRTAG);
			break;

		case CAN_MODE_AUTO_REPLY:
			clearbit(CANCDMOB, CONMOB0);
			setbit(CANCDMOB, CONMOB1);
			setbit(CANCDMOB, RPLV);
			setbit(CANIDT4, RTRTAG);
			break;

		default:
			return 0;
	}

	return 1;
}


/*******************************************************************************
 Get the mode of the previously selected object
*******************************************************************************/
uint8_t can_get_mode(void)
{
	uint8_t mode = 0;

	if (!getbit(CANCDMOB, CONMOB1) && !getbit(CANCDMOB, CONMOB0)) {
		mode = CAN_MODE_DISABLED;
	} else if (!getbit(CANCDMOB, CONMOB1) && getbit(CANCDMOB, CONMOB0) && !getbit(CANIDT4, RTRTAG)) {
		mode = CAN_MODE_TRANSMIT_DATA;
	} else if (!getbit(CANCDMOB, CONMOB1) && getbit(CANCDMOB, CONMOB0) && getbit(CANIDT4, RTRTAG)) {
		mode = CAN_MODE_TRANSMIT_REMOTE;
	} else if (getbit(CANCDMOB, CONMOB1) && !getbit(CANCDMOB, CONMOB0) && !getbit(CANIDT4, RTRTAG)) {
		mode = CAN_MODE_RECEIVE_DATA;
	} else if (getbit(CANCDMOB, CONMOB1) && !getbit(CANCDMOB, CONMOB0) && getbit(CANCDMOB,RPLV) && getbit(CANIDT4, RTRTAG)) {
		mode = CAN_MODE_AUTO_REPLY;
	}

	return mode;
}


/*******************************************************************************
 Writes to the object of the data being transmitted to
*******************************************************************************/
void can_set_data(uint8_t length, uint8_t * data)
{
	uint8_t i;
	uint8_t cancdmob;

	length = Min(length, 8);

	// Number of data bytes in the message apparently is allowed that CANCDMOB 
	// register not arbitrarily write / read many times, so we store the value 
	// in between, delete the corresponding bits for the len, then write usere 
	// len pure as it may, how it works, at least, so caution cleaning up ;-)
	cancdmob = CANCDMOB;
	clearbit(cancdmob, DLC0);
	clearbit(cancdmob, DLC1);
	clearbit(cancdmob, DLC2);
	clearbit(cancdmob, DLC3);
	CANCDMOB = (cancdmob | (length << DLC0));
	for (i = 0; i < length; i++)
	CANMSG = data[i];
}

/*******************************************************************************
 Get received can data and length
*******************************************************************************/
int can_get_data(uint8_t *msg)
{
	uint8_t i;
	int len = can_get_dlc();
	for (i = 0; i < len; i++)
		msg[i] = CANMSG;
	return len;
}

/*******************************************************************************
 Sets the interrupt for the object
 Parameters: uint8_t mob: number of the object (0-14)
*******************************************************************************/
void can_set_mob_interrupt(uint8_t mob)
{
	setbit(CANIE2, iemob[mob]);
}

/*******************************************************************************
 Clears the interrupt according to the subject
 Parameters: uint8_t mob: number of the object (0-14)
*******************************************************************************/
void can_clear_mob_interrupt(uint8_t mob)
{
	clearbit(CANIE2, iemob[mob]);
}

/*******************************************************************************
 Determines which object has triggered interrupt
 Returns: uint8_t mob: number of the object
*******************************************************************************/
uint8_t can_get_mob_interrupt(void)
{
	uint8_t  mob;
	uint16_t maske;
	maske = CANSIT2 | (CANSIT1 << 8);

	// If all 32 bits of the bit mask 0 then an error has occurred
	if (maske == 0)
		return NOMOB;

	// The bit mask is slid so long to the right until Bit 0 a 1 has. 
	// The number of shift operators thus indicates the number of back MOBs
	for (mob = 0; (maske & 0x01) == 0; maske >>= 1, ++mob);

	// Check if mob is greater than the number of available
	// Message object is the result of incorrect
	if (mob > 14)
		return NOMOB;

	return mob;
}

/*******************************************************************************
 Disabled the selected object
*******************************************************************************/
int can_disable_mob(uint8_t mob)
{
	// Select object
	can_get_mob(mob);

	// Enable interrupt for this property
	can_clear_mob_interrupt(mob);

	// Set operating mode
	can_set_mode(CAN_MODE_DISABLED);

	return 1;
}

/*******************************************************************************
 Function sends a request (Remote Frame)
*******************************************************************************/
int can_send_remote(uint8_t mob)
{
	uint8_t mode;

	// select object
	can_get_mob(mob);
  
	// Save current operating mode
	mode = can_get_mode();

	// Start data transfer
	can_set_mode(CAN_MODE_TRANSMIT_REMOTE);

	// Wait is completed until the data transfer (Set TXOK flag of CAN controller)
	while (!getbit(CANSTMOB, TXOK));      

	// Clear TXOK flag by hand
	clearbit(CANSTMOB, TXOK);

	// Restore Old mode
	can_set_mode(mode);

	return 1;
}

void (*interruptHander)() = 0;

/*******************************************************************************
 Interrupt for receiving a message
*******************************************************************************/
SIGNAL(SIG_CAN_INTERRUPT1)
{
  uint8_t save_canpage;
  uint8_t mob;
  //uint32_t id;

  // Save current FANPAGE
  save_canpage = CANPAGE;

  // Index of MOB determine which triggered the interrupt
  mob = can_get_mob_interrupt();
    
  // If it was not a valid MOB Cancel
  if (mob == NOMOB)
    return;

  // Object that triggered the interrupt fetch
  can_get_mob(mob);
  
  // XXX here, what should happen when a interrupt XXX
  // Get Id of the message and read the information from MOBs CANMSG
  if (interruptHander) {
	  interruptHander();
  }

  // Clear RXOK flag
  clearbit(CANSTMOB, RXOK);
    
  // MOB rely on receiving and CAN 2.0B standard
  can_set_mode(CAN_MODE_RECEIVE_DATA);
    
  // Restore CANPAGE
  CANPAGE = save_canpage;
}

/*******************************************************************************
 Set a CAN interrupt handler for user handling logic
 
   for get a CAN data user should add below code on handler
   	id = can_get_id();
   	int len = can_get_data(data); 
*******************************************************************************/
void set_interrupt_handler(void (*handler)(void))
{
	interruptHander = handler;
}