/*
 * AT90CAN_TPMS.cpp
 *
 * Target device: AT90CAN128
 * Created: 2016-01-19 오후 9:21:04
 * Author : lsw2000
 */ 


#include "config.h"	//#define F_CPU 16000000UL
#include <avr/io.h>
#include <util/atomic.h>
#include <util/delay.h>  
#include <stdio.h>
#include <string.h>

#include "MultiSegmentDisplay.h"
//#include "can_lib.h"
#include "can.h"
#include "uart_lib.h"

const uint16_t STEERING_ANGLE_002 = 0x002;
const uint16_t ENGINE_STATUS_180 = 0x180;
const uint16_t ABS_FRONT_284 = 0x284;
const uint16_t CAR_STATUS_358 = 0x358;
const uint16_t TPMS_385 = 0x385;

struct TransmissionInformation {
	uint16_t targetRPM[6];
	uint8_t currentLevel;
};

MultiSegmentDisplay segmentDisplay;
volatile static uint16_t tick_timer = 0;
uint16_t speedx100 = 0;
TransmissionInformation transInfo = { 0, };
bool manualGearNaturalPosition = true;


/*******************************************************************************
 Init. UARTs
*******************************************************************************/
void initUART()
{
	Uart_select(UART_1);
	uart_init(CONF_8BIT_NOPAR_1STOP, 115200);
	uart_mini_printf("\n======= CAN TPMS test =======\n");
}

/*******************************************************************************
 Init. CAN
*******************************************************************************/
void initCAN()
{
	// baudrate: 500K, Interrupts, Std. Id
	if (can_init(CAN_BAUDRATE_500K, CAN_INTERRUPT_RX, 0)) {
		segmentDisplay.setString(0, "CAN OK");
	}
}

/*******************************************************************************
 Get gear position information from current speed and RPMs
*******************************************************************************/
void getTransmissionInformation(TransmissionInformation& info, uint16_t currentRPM, uint16_t currentSpeed)
{	// QM5 manual mission parameters
	const float rpmPerSpeedRatio[] = { 0.00765f, 0.0139f, 0.02157f, 0.03012f, 0.03945f, 0.04785f };
	const float rpmPerSpeedRatioRange[] = { 0.0f, 0.010775f, 0.017735f, 0.025845f, 0.034785f, 0.04365f, 0.1f };

	float calcRatio = (float)currentSpeed / (float)currentRPM;
	
	for (int i = 0; i < 6; i++) {
		// Get current gear level
		if ((rpmPerSpeedRatioRange[i] <= calcRatio) && (calcRatio < rpmPerSpeedRatioRange[i+1])) {
			info.currentLevel = i;
		}
		// Get destination RPM for every gear level for current speed
		info.targetRPM[i] = (int)(currentSpeed / rpmPerSpeedRatio[i]);
	}
}

void getTransmissionInformation2(TransmissionInformation& info, uint16_t currentRPM, uint16_t currentSpeed)
{	// QM5 manual mission parameters
	const float rpmPerSpeedRatio[] = { 765, 1390, 21570, 30120, 39450, 47850 };
	const uint16_t rpmPerSpeedRatioRange[] = { 0, 1077, 1773, 2584, 3478, 4365, 10000 };

	uint32_t x100000CurrentSpeed = currentSpeed * 100000;
	uint16_t calcRatio = x100000CurrentSpeed / currentRPM;
	
	for (int i = 0; i < 6; i++) {
		// Get current gear level
		if ((rpmPerSpeedRatioRange[i] <= calcRatio) && (calcRatio < rpmPerSpeedRatioRange[i+1])) {
			info.currentLevel = i;
		}
		// Get destination RPM for every gear level for current speed
		info.targetRPM[i] = (x100000CurrentSpeed / rpmPerSpeedRatio[i]);
	}
}

/*******************************************************************************
 Display TPMS value
*******************************************************************************/
void showATPMSValue(uint8_t pos, uint16_t value)
{
	static uint16_t prevValue[4] = { 0, 0, 0, 0 };
	if (prevValue[pos/3] != value) {
		prevValue[pos/3] = value;
		uint8_t attr = (value > 20 || value == 0) ? MultiSegmentDisplay::ATTR_BLINK | 4 : // Normal
								MultiSegmentDisplay::ATTR_BLINK | MultiSegmentDisplay::ATTR_BLINK_FAST | 32;
		segmentDisplay.setAttribute(pos, pos + 2, attr);
		//segmentDisplay.setAttribute(pos, pos + 2, MultiSegmentDisplay::ATTR_BLINK | 4);
		//segmentDisplay.setAttribute(pos, pos + 2, MultiSegmentDisplay::ATTR_BLINK | MultiSegmentDisplay::ATTR_BLINK_FAST | 4);
	}
	value ? segmentDisplay.set3DigitValueWithDotPos(pos, value, 1) : segmentDisplay.setString(pos, " - ");
}

/********************************************************************************
 Convert to 3digit value for segment display
********************************************************************************/
void checkTPMSAndDisplay(uint8_t* data)
{
	const float TPMS2PSI_10x = 0.25f * 10.0f;	// Multiply by 10 for int display
	
	uint16_t tirePressFR = data[2] * TPMS2PSI_10x;
	uint16_t tirePressFL = data[3] * TPMS2PSI_10x;
	uint16_t tirePressRR = data[4] * TPMS2PSI_10x;
	uint16_t tirePressRL = data[5] * TPMS2PSI_10x;

	showATPMSValue(0, tirePressFL);
	showATPMSValue(3, tirePressFR);
	showATPMSValue(6, tirePressRL);
	showATPMSValue(9, tirePressRR);
		
	//uart_mini_printf("FL:%d FR:%d RL:%d RR:%d\n", tirePressFL, tirePressFR, tirePressRL, tirePressRR);
}

/*******************************************************************************
 Display Speed
*******************************************************************************/
void displaySpeed(uint8_t* data)
{
	static uint16_t speed = -1;
	speedx100 = (data[4] << 8 | data[5]);
	
	// To avoid frequent changes by the halfway value, No update the top 50% range.
	const uint16_t underPointSpeed = speedx100 % 100;
	if (underPointSpeed < 50) {
		if (speed != speedx100 / 100) {	// Change display buffer only value changed
			speed = speedx100 / 100;
			if (!MultiSegmentDisplay::segmentAttributeBuffer[12]) {		// Update while other item not displaying
				segmentDisplay.set3DigitValueWithDotPos(12, speed, 3);
				segmentDisplay.setAttribute(12, 12 + 2, 0);
			}
		}
	}
}

/*******************************************************************************
 Display steering angle
*******************************************************************************/
void displaySteeringAngle(uint8_t* data)
{
	static uint8_t lastPointPos = 1;
	int16_t steeringAngle = ((data[0]) | ((data[1] & 0xFF) << 8)) / 10;
	bool angleRight = true;
	if (steeringAngle < 0) {
		steeringAngle = -steeringAngle;
		angleRight = false;
	}
	if (angleRight) {
		lastPointPos = MultiSegmentDisplay::blinkDelayCount / 15;
	} else {
		lastPointPos = (60 - MultiSegmentDisplay::blinkDelayCount) / 15;
	}
	if (!MultiSegmentDisplay::segmentAttributeBuffer[12]) {		// Update while other item not displaying
		segmentDisplay.set3DigitValueWithDotPos(12, steeringAngle, lastPointPos);
		//segmentDisplay.setAttribute(12, 12 + 2, MultiSegmentDisplay::ATTR_BLINK_FAST | 1);
	}
}

/*******************************************************************************
 Engine related - Show estimated gear level
*******************************************************************************/
void checkEngineParameter(uint8_t* data)
{
    uint16_t RPM = (data[0] << 7 | data[1]) >> 2;	// Engin RPM = (D1*128+D2)/4
    uint16_t throttle = (data[5] << 2) | ((data[6] & 0xC0) >> 6);
    bool clutchPressed = (data[7] & 0x04);
	
	static bool lastClutchPressed = false;
	static uint8_t validShiftLevel = 0;
	static char outStr[4] = "[0]";
	
	getTransmissionInformation(transInfo, RPM, speedx100 / 100);
	//getTransmissionInformation2(transInfo, RPM, speedx100 / 100);
	if (!clutchPressed) {
		validShiftLevel = transInfo.currentLevel + 1;
	}
	if (!speedx100) {
		validShiftLevel = manualGearNaturalPosition ? 0 : 1;
	}
	outStr[1] = (validShiftLevel + '0');
	
	if (lastClutchPressed != clutchPressed) {
		// Clutch action then display current gear position for a while
		segmentDisplay.setString(12, outStr);
		segmentDisplay.setAttribute(12, 12 + 2, MultiSegmentDisplay::ATTR_BLINK_FAST | 1);
	}
	
	// Update accelerator value
	segmentDisplay.setAccelerator(throttle);
	
	uint16_t shiftDownRpm = 0;
	uint16_t shiftUpRpm = 0;
	//static uint16_t shiftDownRpmOnClutchPressed = 0;
	if (transInfo.currentLevel > 0) {
		shiftDownRpm = transInfo.targetRPM[transInfo.currentLevel - 1];
		//if (clutchPressed) {
			//if (lastClutchPressed != clutchPressed) {
				//shiftDownRpmOnClutchPressed = shiftDownRpm;
			//}
			//shiftDownRpm = shiftDownRpmOnClutchPressed;	// Hold guide RPM value
		//}
	}
	if (transInfo.currentLevel < 5) {
		shiftUpRpm = transInfo.targetRPM[transInfo.currentLevel + 1];
	}
	segmentDisplay.setRpmData(shiftDownRpm, RPM, shiftUpRpm);
	MultiSegmentDisplay::downShiftPointDimming = !clutchPressed;// Prevent flashing if normal driving

	// test
	//segmentDisplay.set3DigitValueWithDotPos(0, throttle, 2);
	
	lastClutchPressed = clutchPressed;
}

/*******************************************************************************
 D2[2] 1 for Manual gear natural position
*******************************************************************************/
void checkCarStatus(uint8_t* data)
{
	manualGearNaturalPosition = (data[1] & 0x04);
}

/*******************************************************************************
 Send CAN data to Bluetooth via serial for logging and monitor
 
 Format for serial sending
 0         1         2
 01234567890123456789012
 245 8 FFE000186310FFE3
*******************************************************************************/
void sendCANtoSerial(uint32_t id, uint8_t len, uint8_t* data)
{
	char buf[24];
	buf[0] = 0;
	sprintf(buf, "%lX %d ", id, len);
	uart_mini_printf(buf);
	//uint8_t netIndex = ((id > 0xff) ? 3 : ((id > 0xf) ? 2: 1)) + 3;
	//for (uint8_t i = 0; i < len; i++) {
		//
	//}
	for (uint8_t i = 0; i < len; i++) {
		buf[0] = 0;
		sprintf(buf, "%02X", data[i]);
		uart_mini_printf(buf);
	}
	uart_mini_printf("\n");
	
}

/*******************************************************************************
*******************************************************************************/
void canInterruptHander() {
	uint32_t id;
	uint8_t data[8];
	
	id = can_get_id();
	can_get_data(data);
	//int len = can_get_data(data);
	
	//segmentDisplay.set3DigitValueWithDotPos(12, id, -1);
	
	switch (id) {
		case TPMS_385 :
			checkTPMSAndDisplay(data);
			break;
			
		case ABS_FRONT_284 :
			displaySpeed(data);
			break;
			
		case STEERING_ANGLE_002 :
			if (speedx100 == 0) {
				displaySteeringAngle(data);
			}
			break;
			
		case ENGINE_STATUS_180 :
			checkEngineParameter(data);
			break;
			
		case CAR_STATUS_358 :
			checkCarStatus(data);
			break;
	}
	
	// + Send CAN data to Bluetooth via serial for logging and monitor for CANTaker(Android app)
	//sendCANtoSerial(id, len, data);
}

/*******************************************************************************
 Show intro display
*******************************************************************************/
void showIntro()
{
	// 
	const char* stringForFrame[] = {
	//   012345678901234
		"             -",
		"            ===",
		" ____  ^^^^    ",
		" ----  ----",
		"^^^^^^______"
		};
	int16_t introAcceleratorValue = 1;
	int16_t introRpmValue = 600;
		
	for (int i = 0; i < 10; i++) {
		segmentDisplay.setString(0, stringForFrame[i%5]);
		
		// setAccelerator
		segmentDisplay.setAccelerator(introAcceleratorValue);
		introAcceleratorValue <<= 1;
		
		// setRpmData
		segmentDisplay.setRpmData(0, introRpmValue, introRpmValue);
		introRpmValue += 600;
		_delay_ms(100);
	}
	
	// Clear accel and rpm
	segmentDisplay.setAccelerator(0);
	segmentDisplay.setRpmData(0, 0, 0);
}

/*******************************************************************************
 MAIN
*******************************************************************************/
int main(void)
{
	// Intro.
	showIntro();
	
	//initUART();	// Debug only - UART operation may delay the segment refresh timer.
	segmentDisplay.setString(0, "CANini");
	initCAN();
	set_interrupt_handler(canInterruptHander);
	
	// mob: 0, modus: senden, id 3, idm 0xffffffff
	//can_enable_mob(0, CAN_MODE_TRANSMIT_DATA, 3, 0xffffffff);
	//can_enable_mob(1, CAN_MODE_RECEIVE_DATA, 0x385, 0xffffffff);
	can_enable_mob(1, CAN_MODE_RECEIVE_DATA, 0x385, 0x0);	// Accept all!
	

	//segmentDisplay.setAttribute(0, 5, MultiSegmentDisplay::ATTR_BLINK | 3);
	
	while (true) {
		
	}
	
	return 1;
}

